// ///////////////////  ***********************  \\\\\\\\\\\\\\\\\\\\ //
//
//  SoSDLRenderArea, version 1.0
//
//  a simple window component like SoXtRenderArea for Inventor
//
//  Copyright (C) 2001-2002 by PC John (pcjohn@email.cz). 
//  All rights reserved.
//
//  This game is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public License
//  version 2.1 as published by the Free Software Foundation. See the
//  file LICENSE.LGPL at the root directory of the distribution for
//  more details.
//
// \\\\\\\\\\\\\\\\\\\  ***********************  //////////////////// //


// Windows specific
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
# include <windows.h>
# undef DELETE // Microsoft defines it and it collides with SoKeyboardEvent::DELETE

// to avoid Microsoft Visual Studio to generate warnings when
// compiling Coin's headers if dll-linking
# pragma warning(disable:4275)
# pragma warning(disable:4251)
#endif

#include <Inventor/SoDB.h>
#include <Inventor/SoInteraction.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFTime.h>
#include <Inventor/elements/SoGLCacheContextElement.h>
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodekits/SoNodeKit.h>
#include <assert.h>
#include <string.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

#include "SoSDLRenderArea.h"


#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#pragma comment(lib, "Opengl32.lib")
#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")
#endif


static char SO_SDL_DEFAULT_WINDOW_TITLE[] = "SoSDLRenderArea";


SoSDLRenderArea* SoSDL::windowsList = NULL;
bool SoSDL::widgetVisible = false;
static SoSDLRenderArea::Resolution *availableRes = NULL;
bool SoSDL::stop = false;
static SDL_TimerID wakeUpTimer = NULL;

static void sensorQueueChanged_CB(void *data);



// Ascii to SoKeyboardEvent::Key translation table for first 128 characters
#define Xx SoKeyboardEvent::ANY // many characters can not be translated
static const SoKeyboardEvent::Key keyTranslateTable[128] =
{
  Xx, Xx, Xx, Xx, Xx, Xx, Xx, Xx, // 0-7
  SoKeyboardEvent::BACKSPACE, // 8
  SoKeyboardEvent::TAB, // 9
  Xx, Xx, Xx, // 10-12
  SoKeyboardEvent::ENTER, // 13
  Xx, Xx, Xx, Xx, Xx, // 14-18
  SoKeyboardEvent::PAUSE, // 19
  Xx, Xx, Xx, Xx, Xx, Xx, Xx, // 20-26
  SoKeyboardEvent::ESCAPE, // 27
  Xx, Xx, // 28-29
  Xx, Xx, // 30-31

  SoKeyboardEvent::SPACE, // 32
  Xx, Xx, Xx, Xx, Xx, Xx, // 33-38
  SoKeyboardEvent::APOSTROPHE, // 39 "'"
  Xx, Xx, Xx, Xx, // 40-43

  SoKeyboardEvent::COMMA, // 44 ","
  SoKeyboardEvent::MINUS, // 45 "-"
  SoKeyboardEvent::PERIOD, // 46 "."
  SoKeyboardEvent::SLASH, // 47 "/"
  
  SoKeyboardEvent::NUMBER_0, // 48
  SoKeyboardEvent::NUMBER_1,
  SoKeyboardEvent::NUMBER_2,
  SoKeyboardEvent::NUMBER_3,
  SoKeyboardEvent::NUMBER_4,
  SoKeyboardEvent::NUMBER_5,
  SoKeyboardEvent::NUMBER_6,
  SoKeyboardEvent::NUMBER_7,
  SoKeyboardEvent::NUMBER_8,
  SoKeyboardEvent::NUMBER_9, // 57

  Xx, // 58
  SoKeyboardEvent::SEMICOLON, // 59 ";"
  Xx, // 60
  SoKeyboardEvent::EQUAL, // 61 "="
  Xx, Xx, Xx, // 62-64

  SoKeyboardEvent::A, // 65 "A"
  SoKeyboardEvent::B,
  SoKeyboardEvent::C,
  SoKeyboardEvent::D,
  SoKeyboardEvent::E,
  SoKeyboardEvent::F,
  SoKeyboardEvent::G,
  SoKeyboardEvent::H,
  SoKeyboardEvent::I,
  SoKeyboardEvent::J,
  SoKeyboardEvent::K,
  SoKeyboardEvent::L,
  SoKeyboardEvent::M,
  SoKeyboardEvent::N,
  SoKeyboardEvent::O,
  SoKeyboardEvent::P,
  SoKeyboardEvent::Q,
  SoKeyboardEvent::R,
  SoKeyboardEvent::S,
  SoKeyboardEvent::T,
  SoKeyboardEvent::U,
  SoKeyboardEvent::V,
  SoKeyboardEvent::W,
  SoKeyboardEvent::X,
  SoKeyboardEvent::Y,
  SoKeyboardEvent::Z, // 90 "Z"
  
  SoKeyboardEvent::BRACKETLEFT, // 91 "["
  SoKeyboardEvent::BACKSLASH, // 92 "\"
  SoKeyboardEvent::BRACKETRIGHT, // 93 "]"
  Xx, Xx, // 94-95
  SoKeyboardEvent::GRAVE, // 96 "`"

  SoKeyboardEvent::A, // 97 "a"
  SoKeyboardEvent::B,
  SoKeyboardEvent::C,
  SoKeyboardEvent::D,
  SoKeyboardEvent::E,
  SoKeyboardEvent::F,
  SoKeyboardEvent::G,
  SoKeyboardEvent::H,
  SoKeyboardEvent::I,
  SoKeyboardEvent::J,
  SoKeyboardEvent::K,
  SoKeyboardEvent::L,
  SoKeyboardEvent::M,
  SoKeyboardEvent::N,
  SoKeyboardEvent::O,
  SoKeyboardEvent::P,
  SoKeyboardEvent::Q,
  SoKeyboardEvent::R,
  SoKeyboardEvent::S,
  SoKeyboardEvent::T,
  SoKeyboardEvent::U,
  SoKeyboardEvent::V,
  SoKeyboardEvent::W,
  SoKeyboardEvent::X,
  SoKeyboardEvent::Y,
  SoKeyboardEvent::Z, // 122 "z"

  Xx, Xx, Xx, Xx, // 123-126

  (SoKeyboardEvent::Key)0xffff // To avoid all troubles with DELETE
                               // key we use this instead of simple
                               // SoKeyboardEvent::DELETE
};

// not assigned keys:
//
// SoKeyboardEvent::PRIOR
// SoKeyboardEvent::NEXT
// SoKeyboardEvent::PAD_F1
// SoKeyboardEvent::PAD_F2
// SoKeyboardEvent::PAD_F3
// SoKeyboardEvent::PAD_F4
// SoKeyboardEvent::PAD_SPACE
// SoKeyboardEvent::PAD_TAB
// SoKeyboardEvent::PAD_INSERT
// SoKeyboardEvent::PAD_DELETE
//
// SoKeyboardEvent::RETURN - same as SoKeyboardEvent::ENTER
//
// SoKeyboardEvent::SHIFT_LOCK


// additional SoKeyboardEvent::Key translation table
static const SoKeyboardEvent::Key specialKeyTranslateTable[] =
{
  SoKeyboardEvent::PAD_0, // 256
  SoKeyboardEvent::PAD_1,
  SoKeyboardEvent::PAD_2,
  SoKeyboardEvent::PAD_3,
  SoKeyboardEvent::PAD_4,
  SoKeyboardEvent::PAD_5,
  SoKeyboardEvent::PAD_6,
  SoKeyboardEvent::PAD_7,
  SoKeyboardEvent::PAD_8,
  SoKeyboardEvent::PAD_9,
  SoKeyboardEvent::PAD_PERIOD,
  SoKeyboardEvent::PAD_DIVIDE,
  SoKeyboardEvent::PAD_MULTIPLY,
  SoKeyboardEvent::PAD_SUBTRACT,
  SoKeyboardEvent::PAD_ADD,
  SoKeyboardEvent::PAD_ENTER, // 271
  Xx,

  SoKeyboardEvent::UP_ARROW, // 273
  SoKeyboardEvent::DOWN_ARROW,
  SoKeyboardEvent::RIGHT_ARROW,
  SoKeyboardEvent::LEFT_ARROW,
  SoKeyboardEvent::INSERT,
  SoKeyboardEvent::HOME,
  SoKeyboardEvent::END,
  SoKeyboardEvent::PAGE_UP,
  SoKeyboardEvent::PAGE_DOWN, // 281

  SoKeyboardEvent::F1, // 282
  SoKeyboardEvent::F2,
  SoKeyboardEvent::F3,
  SoKeyboardEvent::F4,
  SoKeyboardEvent::F5,
  SoKeyboardEvent::F6,
  SoKeyboardEvent::F7,
  SoKeyboardEvent::F8,
  SoKeyboardEvent::F9,
  SoKeyboardEvent::F10,
  SoKeyboardEvent::F11,
  SoKeyboardEvent::F12, // 296

  Xx, Xx, Xx,

  SoKeyboardEvent::NUM_LOCK, // 300
  SoKeyboardEvent::CAPS_LOCK,
  SoKeyboardEvent::SCROLL_LOCK,
  SoKeyboardEvent::RIGHT_SHIFT,
  SoKeyboardEvent::LEFT_SHIFT,
  SoKeyboardEvent::RIGHT_CONTROL,
  SoKeyboardEvent::LEFT_CONTROL,
  SoKeyboardEvent::RIGHT_ALT,
  SoKeyboardEvent::LEFT_ALT, // 308

  Xx, Xx, Xx, Xx, Xx, Xx,
  Xx,

  SoKeyboardEvent::PRINT, // 316

  Xx, Xx, Xx, Xx, Xx, // 317-321
};

#undef Xx



Widget SoSDL::init(const char *appName, const char *className)
{
  // initialize OIV
  SoDB::init();
  SoNodeKit::init();
  SoInteraction::init();

  // initialize SDL
  bool r = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE) == 0;
  if (!r) {
    SoDebugError::post("SoSDL::init",
                       "Can not initialize SDL.");
    stop = true;
    return 0;
  }
  if (atexit(SoSDL::done) != 0) { // this is very neccessary when we are going
    // to switch resolution to be the resolution restored back
    SoDebugError::post("SoSDL::init",
                       "Can not register clean-up atexit function.");
    stop = true;
    return 0;
  }

  // return widget, must be non-zero
  return (Widget) 0xffffffff;
}



void SoSDL::done()
{
  if (availableRes) {
    free(availableRes);
    availableRes = NULL;
  }
  SDL_Quit();
}



void SoSDL::mainLoop()
{
  if (stop)  return;

  SDL_PumpEvents();
  SDL_Event event;

  while (!stop) {
    while (SDL_PollEvent(&event) && !stop) {
    place_event:
      switch (event.type) {
      case SDL_MOUSEMOTION: SoSDLRenderArea::handleMouseMotion(event.motion.x, event.motion.y); break;
      case SDL_MOUSEBUTTONDOWN: SoSDLRenderArea::handleMouseButton(true, event.button.button, event.button.x, event.button.y); break;
      case SDL_MOUSEBUTTONUP: SoSDLRenderArea::handleMouseButton(false, event.button.button, event.button.x, event.button.y); break;
      case SDL_KEYDOWN: SoSDLRenderArea::handleKeyboard(event.key.keysym.sym, true); break;
      case SDL_KEYUP: SoSDLRenderArea::handleKeyboard(event.key.keysym.sym, false); break;
      case SDL_VIDEOEXPOSE: SoSDLRenderArea::handleRender(); break;
      case SDL_VIDEORESIZE: SoSDLRenderArea::handleReshape(event.resize.w, event.resize.h); break;
      case SDL_USEREVENT: break;
      case SDL_QUIT: if (!SoSDLRenderArea::handleClose()) return; break;
			}
		}
                                                                  
    if (processSensors() == false && !stop) {
      SoDB::getSensorManager()->setChangedCallback(sensorQueueChangedCallback, NULL);
      if (SDL_WaitEvent(&event) != 1) {
        SoDebugError::post("SoSDL::mainLoop",
                           "SDL_WaitEvent in mainLoop failed.");
        return; // leave mainLoop
      }
      SoDB::getSensorManager()->setChangedCallback(NULL, NULL);
      goto place_event;
    }
	}
}



void SoSDL::show(Widget widget)
// shows all windows of application with visible set to true
{
  if (SoSDL::widgetVisible)  return;

  SoSDL::widgetVisible = true;

  SoSDLRenderArea *w = SoSDL::windowsList;
  while (w != NULL)
  {
    if (w->visible == true)
      w->doShow();
    w = w->nextWindowsListItem;
  }
}



void SoSDL::hide(Widget widget)
// hides all application windows
{
  if (SoSDL::widgetVisible == false)  return;

  SoSDL::widgetVisible = false;

  SoSDLRenderArea *w = SoSDL::windowsList;
  while (w != NULL)
  {
    if (w->visible == true)
      w->doHide();
    w = w->nextWindowsListItem;
  }
}



void SoSDL::addWindow(SoSDLRenderArea *window)
// adds window to the end of the list
{
  SoSDLRenderArea *w = windowsList;

  if (w==NULL)
  {
    windowsList = window;
    window->nextWindowsListItem = NULL;
    return;
  }

  while (w->nextWindowsListItem != NULL)
    w = w->nextWindowsListItem;

  w->nextWindowsListItem = window;
  window->nextWindowsListItem = NULL;
}



void SoSDL::removeWindow(SoSDLRenderArea *window)
{
  SoSDLRenderArea **w = &windowsList;

  while ((*w) != NULL)
  {
    if ((*w) == window)
    {
      (*w) = (*w)->nextWindowsListItem;
      break;
    }
    w = &(*w)->nextWindowsListItem;
  }
}


  
SoSDLRenderArea* SoSDL::getWindowFromHandle(Widget id)
{
  SoSDLRenderArea *w = windowsList;

  while (w != NULL)
  {
    if (w->handle == id)
      return w;

    w = w->nextWindowsListItem;
  }
  
  return NULL;
}



void SoSDL::sensorQueueChangedCallback(void *data)
{
  postWakeUpMessageCallback(0, NULL);
}



unsigned int SoSDL::postWakeUpMessageCallback(unsigned int interval, void *userData)
{
  if (wakeUpTimer) {
    if (!SDL_RemoveTimer(wakeUpTimer))
      SoDebugError::post("SoSDL::postWakeUpMessageCallback",
                         "Can not remove timer.");
    wakeUpTimer = NULL;
  }

  SDL_Event event;
  event.type = SDL_USEREVENT;
  event.user.code = 0;
  event.user.data1 = NULL;
  event.user.data2 = NULL;
  if (SDL_PushEvent(&event), false) { // FIXME: SDL_PushEvent result should be compared
      // to be nonzero by documentation to get error information. But this gives some
      // strange behaviour (on Windows must we compare to one instead of zero?, on Linux
      // return value looks like random one :-) ), so we will do not check for error here.
      // PCJohn 2002-04-08
    SoDebugError::post("SoSDL::postWakeUpMessageCallback",
                       "SDL_PushEvent failed.");
    return 0; // this will result in application sleeping until external event
  }
  return 0;
}



bool SoSDL::processSensors(void)
{
  SoSensorManager *sm = SoDB::getSensorManager();

  sm->processTimerQueue();
  sm->processDelayQueue(TRUE);

  if (sm->isDelaySensorPending()) {
    return true; 
  } else {
    SbTime eventTime;
    if (sm->isTimerSensorPending(eventTime)) { 
      // only timed events remains
      int t = int((eventTime.getValue() - SbTime::getTimeOfDay().getValue()) * 1000);
      unsigned int ut = (t<=0) ? 0 : (unsigned int)t;
      if (wakeUpTimer != NULL)
        if (!SDL_RemoveTimer(wakeUpTimer))
          SoDebugError::post("SoSDL::processSensors",
                             "Can not remove timer.");
      wakeUpTimer = SDL_AddTimer(ut, postWakeUpMessageCallback, NULL);
      if (wakeUpTimer == NULL) {
        SoDebugError::post("SoSDL::processSensors",
                           "SDL_AddTimer failed.");
        return true; // true means application will not sleep
      }
      return false;
    } else
      // no event remains
      return false; 
  }
}




/////////////////////////////////////////////////////////////////////




SoSDLRenderArea::SoSDLRenderArea(Widget parent,
    const char *name,
    SbBool buildInsideParent,
    SbBool getMouseInput,
    SbBool getKeyboardInput)
{
  handle = NULL;
  winSize = SbVec2s(300, 300);
  title = SO_SDL_DEFAULT_WINDOW_TITLE;
  iconTitle = SO_SDL_DEFAULT_WINDOW_TITLE;
  doubleBuffer = true;
  border = true;

  acceptMouseInput = getMouseInput;
  acceptKeyboardInput = getKeyboardInput;
  mouseX = -1;
  mouseY = -1;

  visible = false;
  fullScreen = false;
  fsWidth = 640;
  fsHeight = 480;
  fsBPP = 0;
  position = SbVec2s(0, 0);
  clearColorBuffer = true;
  clearZBuffer = true;
  autoRedraw = true;
  closeCallback = NULL;

  SoSDL::addWindow(this);

  sceneManager = new SoSceneManager;                            
  ownedSceneManager = true;
  sceneManager->setRenderCallback(sceneManagerRenderCallback, (void *)this);
  sizeChanged(winSize[0], winSize[1]);
}



SoSDLRenderArea::~SoSDLRenderArea()
{
  if (title && title != SO_SDL_DEFAULT_WINDOW_TITLE)  free(title);
  if (iconTitle && iconTitle != SO_SDL_DEFAULT_WINDOW_TITLE)  free(iconTitle);
  if (ownedSceneManager) delete sceneManager;
  destroyWindow();
  SoSDL::removeWindow(this);
}



void SoSDLRenderArea::recreateWindow(SbVec2s size)
{
  handle = SDL_SetVideoMode(size[0], size[1], (fullScreen ? fsBPP : 0), 
      SDL_OPENGL | (fullScreen ? SDL_FULLSCREEN : SDL_RESIZABLE) |
      (border ? 0 : SDL_NOFRAME));
  if (handle == NULL) {
    SoDebugError::post("SoSDLRenderArea::recreateWindow",
                       "Can not create window (width: %i, height: %i, fullScreen: %s, fullScreenBPP: %i, border: %s).",
                       size[0], size[1], (fullScreen ? "true" : "false"), fsBPP, (border ? "true" : "false"));
    SoSDL::stop = true;
    return;
  }

#ifdef __COIN__
  unsigned int cacheContext = SoGLCacheContextElement::getUniqueCacheContext();
#else // SGI or TGS Inventor
  static uint32_t uniqcc = 19720408;
  cacheContext = uniqcc++;
#endif
  sceneManager->getGLRenderAction()->setCacheContext(cacheContext);
}



void SoSDLRenderArea::createWindow()
{
  if (handle == NULL) {

    if (SDL_InitSubSystem(SDL_INIT_VIDEO) != 0) {
      SoDebugError::post("SoSDLRenderArea::createWindow",
                         "SDL video init failed.");
      SoSDL::stop = true;
      return;
    }

    assert(SDL_GetVideoSurface() == NULL && "Only one window can be showed at a time (SDL limitation).");

    SDL_WM_SetCaption(title, iconTitle);
    
    if (SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, (doubleBuffer) ? 1 : 0) != 0) {
      SoDebugError::post("SoSDLRenderArea::createWindow",
                         "Can not set double buffer attribute.");
      SoSDL::stop = true;
      return;
    }

    // window creation
    SbVec2s size;
    if (fullScreen) size = SbVec2s(fsWidth, fsHeight);
    else size = winSize;
    recreateWindow(size);
    sizeChanged(size[0], size[1]);

    if (acceptKeyboardInput) {
      if (SDL_EnableKeyRepeat(0, 0) != 0) {
        SoDebugError::postWarning("SoSDLRenderArea::createWindow",
                                  "Can not disable key repeat.");
      }
    }
  }
}



void SoSDLRenderArea::destroyWindow()
{
  if (handle != NULL) {
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
    handle = NULL;
  }
}



void SoSDLRenderArea::show()
{
  if (visible == true)  return;

  visible = true;

  if (SoSDL::widgetVisible)
    doShow();
}



void SoSDLRenderArea::doShow()
{
  if (handle == NULL) {
    createWindow();
    sceneManager->activate();
  }
};



void SoSDLRenderArea::hide()
{
  if (visible == false)  return;

  visible = false;

  if (SoSDL::widgetVisible)
    doHide();
}



void SoSDLRenderArea::doHide()
{
  if (handle != NULL) {
    destroyWindow();
    sceneManager->deactivate();
  }
};



void SoSDLRenderArea::setSceneManager(SoSceneManager *sm)
{
  assert(sm != NULL);

  if (sceneManager == sm) return;

  sm->setWindowSize(sceneManager->getWindowSize());

  if (ownedSceneManager) {
    delete sceneManager;
    ownedSceneManager = false;
  } else
    sceneManager->setRenderCallback(NULL, NULL);
  unsigned int cacheContext = sceneManager->getGLRenderAction()->getCacheContext();

  sceneManager = sm;
  sceneManager->setRenderCallback(sceneManagerRenderCallback, (void *)this);
  sceneManager->getGLRenderAction()->setCacheContext(cacheContext);
  if (this->isVisible()) sceneManager->activate();
  else sceneManager->deactivate();
}



void SoSDLRenderArea::setTitle(const char *newTitle)
{
  if (title == NULL)
    if (newTitle == NULL) return;
    else;
  else
    if (newTitle == NULL) {
      free(title);
      title = NULL;
      goto update;
    } else
      if (strcmp(title, newTitle) == 0) return;
      else 
        if (title != SO_SDL_DEFAULT_WINDOW_TITLE) free(title);

  title = (char*)malloc(strlen(newTitle)+1);
  strcpy(title, newTitle);
update:
  if (handle != NULL && !fullScreen) {
    SDL_WM_SetCaption(title, iconTitle);
  }
};



void SoSDLRenderArea::setIconTitle(const char *newTitle)
{
  if (iconTitle == NULL)
    if (newTitle == NULL) return;
    else;
  else
    if (newTitle == NULL) {
      free(iconTitle);
      iconTitle = NULL;
      goto update;
    } else
      if (strcmp(iconTitle, newTitle) == 0) return;
      else 
        if (iconTitle != SO_SDL_DEFAULT_WINDOW_TITLE) free(iconTitle);

  iconTitle = (char*)malloc(strlen(newTitle)+1);
  strcpy(iconTitle, newTitle);
update:
  if (handle != NULL && !fullScreen) {
    SDL_WM_SetCaption(title, iconTitle);
  }
}



void SoSDLRenderArea::setDoubleBuffer(SbBool onOrOff)
{
  if (handle != NULL) { // switching can be done only before windows creation
    SoDebugError::postWarning("SoSDLRenderArea::setDoubleBuffer",
                              "Double buffering setting can not be changed after window creation.");
    return; 
  }
  doubleBuffer = (onOrOff == TRUE);
}



void SoSDLRenderArea::setBorder(SbBool onOrOff)
{
  if (handle != NULL) { // switching can be done only before windows creation
    SoDebugError::postWarning("SoSDLRenderArea::setBorder",
                              "Border setting can not be changed after window creation.");
    return; 
  }
  border = (onOrOff == TRUE);
}

  
  
/*
  FIXME: How to set window position with SDL?  20020308 PCJohn
void SoSDLRenderArea::setPosition(SbVec2s pos)
{
  if (fullScreen)  return; // can not set position on full screen window

  position = pos;
}



SbVec2s SoSDLRenderArea::getPosition()
{
  if (fullScreen)
    return SbVec2s(0, 0);

  return position;
}*/



void SoSDLRenderArea::setSize(const SbVec2s &size)
{
  if (isVisible()) 
    recreateWindow(size);
  sizeChanged(size[0], size[1]);
};



SbVec2s SoSDLRenderArea::getSize() const
{
  return sceneManager->getWindowSize();
}



void SoSDLRenderArea::sizeChanged(int w, int h)
{
  sceneManager->setWindowSize(SbVec2s(w, h));

  if (fullScreen) {
    fsWidth = w;
    fsHeight = h;
  } else
    winSize = SbVec2s(w, h);

  renderCallback();
}



void SoSDLRenderArea::handleReshape(int w, int h)
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle");
  ra->setSize(SbVec2s(w, h));
}



void SoSDLRenderArea::setViewportRegion(const SbViewportRegion &newRegion)
{
  sceneManager->setViewportRegion(newRegion);
};



const SbViewportRegion& SoSDLRenderArea::getViewportRegion() const
{
  return sceneManager->getViewportRegion();
}

  
  
void SoSDLRenderArea::setTransparencyType(SoGLRenderAction::TransparencyType type)
{
  sceneManager->getGLRenderAction()->setTransparencyType(type);
}



SoGLRenderAction::TransparencyType SoSDLRenderArea::getTransparencyType() const
{
  return sceneManager->getGLRenderAction()->getTransparencyType();
}



void SoSDLRenderArea::renderCallback()
{
  if (handle != NULL) {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    sceneManager->render(clearColorBuffer, clearZBuffer);

    SDL_GL_SwapBuffers();
  }
}



void SoSDLRenderArea::handleRender()
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle");
  ra->renderCallback();
}



void SoSDLRenderArea::sceneManagerRenderCallback(void *userData, SoSceneManager *mgr)
{
  SoSDLRenderArea *ra = (SoSDLRenderArea*)userData;
  if (ra->autoRedraw)
    ra->renderCallback();
}



void SoSDLRenderArea::makeEvent(SoEvent *event)
{
  // set whole event stuff
  event->setTime(SbTime::getTimeOfDay());
  event->setPosition(SbVec2s(mouseX, this->getSize()[1] - mouseY - 1));

  SDLMod mod = SDL_GetModState();
  event->setShiftDown(mod & (KMOD_LSHIFT | KMOD_RSHIFT));
  event->setCtrlDown(mod & (KMOD_LCTRL | KMOD_RCTRL));
  event->setAltDown(mod & (KMOD_LALT | KMOD_RALT));

  // process timer queue before event
  SoDB::getSensorManager()->processTimerQueue();

  // process event
  sceneManager->processEvent(event);
}



void SoSDLRenderArea::handleKeyboard(int key, bool down)
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle.");

  if (!ra->acceptKeyboardInput)  return;
  
  static SoKeyboardEvent keyEvent;
  SoKeyboardEvent::Key tk;

  if (key<128)  
    if (key>=0)  tk = keyTranslateTable[key];
    else  tk = SoKeyboardEvent::ANY;
  else
    if (key>=256)
      if (key<=321)  tk = specialKeyTranslateTable[key-256];
      else  tk = SoKeyboardEvent::ANY;
    else  tk = SoKeyboardEvent::ANY;
      
  keyEvent.setKey(tk);
  keyEvent.setState(down ? SoButtonEvent::DOWN : SoButtonEvent::UP);

  ra->makeEvent(&keyEvent);
}



void SoSDLRenderArea::handleMouseMotion(int x, int y)
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle.");

  if (!ra->acceptMouseInput)  return;

  if (x != ra->mouseX || y != ra->mouseY) {
    static SoLocation2Event mouseLocationEvent;
    ra->mouseX = x;
    ra->mouseY = y;
    ra->makeEvent(&mouseLocationEvent);
  }
}



void SoSDLRenderArea::handleMouseButton(bool down, int button, int x, int y)
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle.");

  if (!ra->acceptMouseInput)  return;

  ra->mouseX = x;
  ra->mouseY = y;
  SoMouseButtonEvent::Button b;
  switch (button) {
  case SDL_BUTTON_LEFT: b = SoMouseButtonEvent::BUTTON1; break;
  case SDL_BUTTON_RIGHT: b = SoMouseButtonEvent::BUTTON2; break;
  case SDL_BUTTON_MIDDLE: b = SoMouseButtonEvent::BUTTON3; break;
  default: b = SoMouseButtonEvent::ANY; break;
  }
  static SoMouseButtonEvent mouseButtonEvent;
  mouseButtonEvent.setState(down ? SoButtonEvent::DOWN : SoButtonEvent::UP);
  mouseButtonEvent.setButton(b);
  ra->makeEvent(&mouseButtonEvent);
}



bool SoSDLRenderArea::handleClose()
{
  SoSDLRenderArea *ra = SoSDLRenderArea::getComponent(SDL_GetVideoSurface());
  assert(ra != NULL && "invalid window handle.");

  // treat window as it have been hiden
  ra->handle = NULL;
  ra->visible = false;

  // call close callback
  bool reg = (ra->closeCallback != NULL);
  if (reg)
    ra->closeCallback(ra->closeCBData, ra);

  return reg;
}



void SoSDLRenderArea::setFullScreen(bool b)
{
  if (fullScreen != b) {
    destroyWindow();

    fullScreen = b;
    if (!fullScreen)
      sceneManager->setWindowSize(winSize);
    if (visible)
      doShow();
  }
}
                           


#define DEPTH_NUM 3
static struct {
  int depth;
  SoSDLRenderArea::Resolution *res;
  int num;
} depthList[DEPTH_NUM] = 
{
  { 16, NULL, 0 },
  { 24, NULL, 0 },
  { 32, NULL, 0 },
};
static int resolutionNum;



static void updateResolutionList()
{
  if (availableRes)  return;

  int i;
  int num;
  int sum = 0;;

  SDL_PixelFormat format;
  SDL_Rect **resl;
  memset(&format, sizeof(format), 0);
  for (i=0; i<DEPTH_NUM; i++) {
    format.BitsPerPixel = depthList[i].depth;
    resl = SDL_ListModes(&format, SDL_OPENGL | SDL_FULLSCREEN);
    assert(resl != (SDL_Rect **)-1); 
    if (resl != (SDL_Rect **)0 && resl != (SDL_Rect **)-1) {
      num = 0;    
      while (resl[num]) num++;
      sum += num;
    }
  }

  resolutionNum = sum;
  if (sum == 0) {
    availableRes = NULL;
    for (i=0; i<DEPTH_NUM; i++) {
      depthList[i].res = NULL;
      depthList[i].num = 0;
    }
    return;
  }

  availableRes = (SoSDLRenderArea::Resolution*)malloc(sum*sizeof(SoSDLRenderArea::Resolution));

  sum = 0;
  for (i=0; i<DEPTH_NUM; i++) {
    depthList[i].res = &availableRes[sum];
    format.BitsPerPixel = depthList[i].depth;
    resl = SDL_ListModes(&format, SDL_OPENGL | SDL_FULLSCREEN);
    assert(resl != (SDL_Rect **)-1); 
    if (resl == (SDL_Rect **)0 || resl == (SDL_Rect **)-1) {
      depthList[i].res = NULL;
      depthList[i].num = 0;
    } else {
      num = 0;
      while (resl[num]) {
        availableRes[sum].width  = resl[num]->w;
        availableRes[sum].height = resl[num]->h;
        availableRes[sum].bpp    = depthList[i].depth;
        sum++;
        num++;
      }
      depthList[i].num = num;
    }
  }
}



int SoSDLRenderArea::getResolutionList(const Resolution *&resList, int depthBits)
{
  updateResolutionList();

  if (depthBits == 0) {
    resList = availableRes;
    return resolutionNum;
  } else {
    int i;
    for (i=0; i<DEPTH_NUM; i++) {
      if (depthList[i].depth == depthBits) {
        resList = depthList[i].res;
        return depthList[i].num;
      }
    }
    resList = NULL;
    return 0;
  }
}



bool SoSDLRenderArea::isResolutionSupported(int width, int height, int bpp)
{
  if (bpp == 0)
    return (SDL_VideoModeOK(width, height, bpp, SDL_OPENGL | SDL_FULLSCREEN) != 0);
  else
    return (SDL_VideoModeOK(width, height, bpp, SDL_OPENGL | SDL_FULLSCREEN) == bpp);
}



