// ///////////////////  ***********************  \\\\\\\\\\\\\\\\\\\\ //
//
//  This simple example demonstrates usage of SoGlut and SoSDL RenderArea.
//
//  Created by PC John (pcjohn@email.cz), 2001-2002, and 
//  placed to public domain.
//
// \\\\\\\\\\\\\\\\\\\  ***********************  //////////////////// //


#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
// to avoid Microsoft Visual Studio to generate warnings when
// compiling Coin's headers if dll-linking
# pragma warning(disable:4275)
# pragma warning(disable:4251)
#endif


#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/errors/SoError.h>
#include <Inventor/nodes/SoText2.h>
#include <stdio.h>
#include "SoSDLRenderArea.h"


#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
# include <windows.h>
# pragma comment(lib,"user32.lib") // we need this for MessageBox
# pragma comment(linker,"/entry:\"mainCRTStartup\"") // this pragma 
      // is used to change default linking from WinMain
      // to main on Windows platform
# if defined(__COIN__) 
#  pragma comment(lib, "coin1d.lib") // linking with Inventor library in MSVC,
    // for TGS or SGI Inventor you must link manualy propriate libraries
# endif
#endif


/////////////////////////////////////////////////////////////////////


// terminate application on press of any key
void kbFunc(void *userData, SoEventCallback *node)
{
  const SoKeyboardEvent *event = (SoKeyboardEvent*)node->getEvent();
  SoXtRenderArea *myRenderArea = (SoXtRenderArea*)userData;
  
  if (event->getState() == SoButtonEvent::DOWN) {
    switch (event->getKey()) {
    case SoKeyboardEvent::ESCAPE:
      exit(EXIT_SUCCESS);
      break;
    case SoKeyboardEvent::NUMBER_6:
      myRenderArea->setSize(SbVec2s(640,480));
      break;
    case SoKeyboardEvent::NUMBER_8:
      myRenderArea->setSize(SbVec2s(800,600));
      break;
    case SoKeyboardEvent::NUMBER_1:
      myRenderArea->setSize(SbVec2s(1024,768));
      break;
    case SoKeyboardEvent::NUMBER_2:
      myRenderArea->setSize(SbVec2s(1280,1024));
      break;
    default:
      myRenderArea->setFullScreen(!myRenderArea->isFullScreen());
    }
  }
}


// error handler
void error_cb(const SoError *error, void *data)
{
  const char *msg = error->getDebugString().getString();
#ifdef _WIN32
  MessageBox(NULL, msg, "RenderArea error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
#else
  printf("RenderArea error: %s\n", msg);
#endif
}

  
// main function
int main(int argc, char ** argv)
{
  // initialize Inventor and So???RenderArea
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == (void*)NULL)
    exit(1);

  SoError::setHandlerCallback(error_cb, NULL);
	
  // create scene graph
  SoSeparator *root = new SoSeparator;
  root->ref();

  SoPerspectiveCamera *camera = new SoPerspectiveCamera;
  camera->position.setValue(0.f, 0.f, 5.f);
  root->addChild(camera);

  root->addChild(new SoDirectionalLight);

  SoMaterial *material = new SoMaterial;
  material->diffuseColor.setValue(1.f, 0.f, 0.f);
  root->addChild(material);

  root->addChild(new SoCone);

  SoOrthographicCamera *camera2 = new SoOrthographicCamera;
  camera2->height = 100;
  camera2->position.setValue(46.5f, -45.f, 1.f);
  root->addChild(camera2);

  SoText2 *text = new SoText2;
  text->string.set1Value(0, "SoSDLRenderArea Hello cone example");
  text->string.set1Value(1, "");
  text->string.set1Value(2, "Keys:");
  text->string.set1Value(3, "any key to toggle fullscreen/window");
  text->string.set1Value(4, "ESC for exit,");
  text->string.set1Value(5, "6 for setting size to 640x480");
  text->string.set1Value(6, "8 for setting size to 800x600");
  text->string.set1Value(7, "1 for setting size to 1024x768");
  text->string.set1Value(8, "2 for setting size to 1280x1024");
  text->string.set1Value(9, "");
  text->string.set1Value(10, "Supported resolutions:");
  root->addChild(text);

  const SoXtRenderArea::Resolution* resList;
#ifndef __APPLE__ // on current Apple's SDL implementation program crashes here
  int count = SoXtRenderArea::getResolutionList(resList);
#else
  int count = 0;
#endif
  for (int i=0; i<count; i++) {
    char buf[20];
    sprintf(buf, "%dx%d:%d\n", resList[i].width, resList[i].height, resList[i].bpp);
    text->string.set1Value(i+11, buf);
  }
  if (count == 0) {
    text->string.set1Value(11, "no resolution detected or incomplete");
    text->string.set1Value(12, "SDL implementation (this applies");
    text->string.set1Value(13, "to Mac OS X)");
  }

  // create render area
  SoXtRenderArea *myRenderArea = new SoXtRenderArea(myWindow);

  myRenderArea->setBackgroundColor(SbColor(0.2f, 0.2f, 0.2f));
  myRenderArea->setSceneGraph(root);
  myRenderArea->show();


  // set fullscreen, command line parameters must be something
  // like this: 800x600:16 (width x height : bitsPerPixel)
  if (argc >= 2) {
    int width, height, bpp;
    if (sscanf(argv[1], "%dx%d:%d", &width, &height, &bpp) == 3) {
      myRenderArea->setFullScreenParams(width, height, bpp);
      myRenderArea->setFullScreen(true);
    }
  } else
    myRenderArea->setFullScreenParams(800, 600, 16);

  // for keyboard input
  SoEventCallback *eventCB = new SoEventCallback;
  eventCB->addEventCallback(SoKeyboardEvent::getClassTypeId(), kbFunc, myRenderArea);
  root->addChild(eventCB);


  SoXt::show(myWindow);
  SoXt::mainLoop();

  return 0;

}

